--
LIBRARY ieee;
USE ieee.std_logic_1164.all;
USE ieee.std_logic_arith.all;
LIBRARY Cosmix_lib;
USE Cosmix_lib.Cosmix.all;

ENTITY spi_coinc IS
   GENERIC( 
      ncoinc : integer := 6
   );
   PORT( 
      reset : IN     std_logic;
      ck    : IN     std_logic;
      SPDI  : IN     std_logic;
      SPCSn : IN     std_logic;
      SPDO  : OUT    std_logic;
      pp    : OUT    slv4array (ncoinc-1 DOWNTO 0)
   );

-- Declarations

END spi_coinc ;

--
ARCHITECTURE v0 OF spi_coinc IS
 --** Synchronous version **
   type StateType is (StandBy,GetArg,GetPattern,GetDummy,NotMe);
   signal MCState : StateType;
   signal SerialReg: std_logic_vector (7 downto 0) ;
   signal BitCount: integer range 0 to 7; -- Counter for use in State M/C
   signal Arg: integer range 0 to 255 ;
  
BEGIN
   
--

-- Read from master only in this implementation so SPDO High-Z
-- The logic assumes CKPOL=0, CKPHA=0 in the C8051 SPI clock config.
  
  SPDO <= 'Z';

-- Use the rising edge to update the bit counter
  
  process (ck,reset,SPCSn)
   begin
      if((reset = '1') or (SPCSn = '1') ) then
         BitCount <= 0;
      elsif (ck'event and ck='1') then
         if (BitCount = 7) then
           BitCount <= 0;
         else
           BitCount <= BitCount + 1;
         end if;
      end if;
   end process;

-- Rising edge should be centered in data
   
   process (ck,reset,SPCSn)
   begin

     if((reset = '1') or (SPCSn = '1') ) then

        MCState <= StandBy;
        SerialReg <= (others => '0');
        Arg <= 0;

      elsif( ck'event and ck='1' ) then  -- rising edge

         SerialReg <= SerialReg(6 downto 0) & SPDI ; -- shift in next bit

         case MCState is
           
-- First byte is the OpCode
         when StandBy =>
            if(BitCount = 7) then
               if ((SerialReg(6 DOWNTO 0) & SPDI) = x"88" ) then
                  MCState <= GetArg;
               else
                  MCState <= Notme;
               end if ;
            end if ;

-- Second byte (OpArg) contains the target coincidence channel index
         when GetArg =>
            if(BitCount = 7) then
               Arg <= conv_integer(unsigned(SerialReg(6 DOWNTO 0) & SPDI));
               MCState <= GetPattern;
            end if;

-- Third byte contains the coincidence pattern in the 4 LSB
         when GetPattern =>
            if(BitCount = 7) then
               pp(Arg) <= std_logic_vector(SerialReg(2 DOWNTO 0) & SPDI);
               MCState <= GetDummy;
            end if;
            
-- Fourth byte is dummy
         when GetDummy =>
            if(BitCount = 7) then
               MCState <= StandBy;
            end if;
--
         when NotMe =>                  -- Sit here until reset
           Arg <= 0;
--            
         end case ;
      end if ;      
   end process;
--    
END v0;

