----------------------------------------------------------------------------------
-- CROCS - AsyncSerialRX
-- Async serial in to parallel data out with newData pulse after each new byte.
--
-- Fergus Noble 2011
----------------------------------------------------------------------------------

LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.ALL;

LIBRARY work;
USE work.Globals.ALL;

entity AsyncSerialRX is
	 Generic ( BaudRate : integer := 9600 );
    Port ( serialIn : in  STD_LOGIC;
           clk : in  STD_LOGIC;
           rst : in  STD_LOGIC;
           newData : out  STD_LOGIC;	-- high whilst dataOut is valid.
           dataOut : out  STD_LOGIC_VECTOR (7 downto 0));
end AsyncSerialRX;

architecture Behavioral of AsyncSerialRX is
 
	type StateType is (idle, start_wait, get_bit, stop);
	signal state : StateType;
	signal timer : integer range 0 to 65535; -- minimum baud rate therefore about 800 baud.
	signal bit_num : integer range 0 to 7;
	signal data : std_logic_vector (7 downto 0);
	
	--constant bit_period : time := 1 sec / BaudRate;
	constant clocks_per_bit : integer := (clk_freq / BaudRate) - 1; -- (-1 since we are counting from zero)
	
begin
	state_machine: process (clk, rst)
	begin
		if (rst = '1') then
			-- async reset
			timer <= 0;
			bit_num <= 0;
			data <= (others => '0');
			state <= idle;
		elsif (clk'event and clk = '1') then
		
			timer <= timer + 1;
			
			case state is
				
				when idle =>					
					-- Wait for start bit, i.e. for serialIn to go low
					if (serialIn = '0') then
						timer <= 0;	-- rst the timer for start_wait
						state <= start_wait;
					else
						state <= idle;
					end if;
					
				when start_wait =>
					-- Wait for the middle of the start bit and sample again to
					-- make sure its still low.
					if (timer >= clocks_per_bit / 2) then
						if (serialIn = '0') then
							-- If serialIn is still low in the middle of the start bit 
							-- period then go ahead and read in the bits.
							timer <= 0;
							bit_num <= 0;
							state <= get_bit;
							-- Set newData to false as we are about to clobber dataOut
							newData <= '0';
						else
							-- Probably a glitch, go back to idle.
							state <= idle;
						end if;
					else
						state <= start_wait;
					end if;
					
				when get_bit =>
					-- Wait for a bit period and read in a bit.
					if (timer >= clocks_per_bit) then
						-- ok, now read in the bit
						data(bit_num) <= serialIn;
						timer <= 0;
						if (bit_num = 7) then
							-- If we have enough bits then check the stop bit.
							state <= stop;
						else
							-- Else carry on reading in bits.
							bit_num <= bit_num + 1;
							state <= get_bit;
						end if;
					else
						state <= get_bit;
					end if;
					
				when stop =>
					-- Check the stop bit is ok (i.e. serialIn high) and if
					-- so set dataOut and newData.
					if (timer >= clocks_per_bit) then
						-- ok, now check the stop bit
						if (serialIn = '1') then
							-- stop bit ok
							dataOut <= data;
							newData <= '1';
							state <= idle;
						else
							-- bad stop bit, discard byte and go back to idle.
							state <= idle;
						end if;
					else
						state <= stop;
					end if;					
					
			end case;
		end if;
	end process state_machine;
end Behavioral; -- AsyncSerialRX

