----------------------------------------------------------------------------------
-- CROCS - GPSTime
-- Main timing module, takes in pps signal from the GPS and hit signal and
-- writes hit times into a packet buffer. Also reads NMEA messages from an
-- NMEA decoder block and writes the appropriate info into the packet buffer.
--
-- Fergus Noble 2011
----------------------------------------------------------------------------------

LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.ALL;

LIBRARY work;
USE work.Globals.ALL;

entity GPSTime is
	 Generic ( ADDR_WIDTH : integer := 15 );
    Port ( nmeaMsgAddr : out STD_LOGIC_VECTOR (7 downto 0);
			  nmeaMsgData : in STD_LOGIC_VECTOR (7 downto 0);
			  nmeaMsgValid : in STD_LOGIC;
           pps : in STD_LOGIC;
			  
           clk : in STD_LOGIC;
           rst : in STD_LOGIC;
			  
			  hit : in STD_LOGIC_VECTOR (3 downto 0);
			  
           addr : in std_logic_vector (ADDR_WIDTH downto 0);
			  data : out std_logic_vector (7 downto 0);
			  newData : out std_logic
			);
end GPSTime;

architecture Behavioral of GPSTime is
	signal buff_addr : unsigned (ADDR_WIDTH downto 0);
	signal buff_data : std_logic_vector (7 downto 0);
	signal buff_page : std_logic;
	signal buff_write_enable : std_logic;
	signal packet_buffer : GPSTimePacketType;

	-- pps process
	type ppsStateType is (
		pps_idle,
		pps_write_final_count_1,
		pps_write_final_count_2,
		pps_write_final_count_3,
		pps_write_final_count_4,
		pps_write_packet_num_1,
		pps_write_packet_num_2,
		pps_write_hit_count_1,
		pps_write_hit_count_2,
		pps_newData,
		pps_newData_2
	);
	signal pps_state : ppsStateType;
	signal pps_previous : std_logic;
	signal pps_buff_addr : unsigned (ADDR_WIDTH downto 0);
	signal pps_buff_data : std_logic_vector (7 downto 0);
	signal pps_wants_bus : std_logic;
	signal timer : unsigned (27 downto 0);	-- Main timer, counts up at clock freq, rst on pps.
	signal timer_final : std_logic_vector (27 downto 0); -- timer final value on rising edge of pps.
	signal packet_number : unsigned (15 downto 0); -- Allows us to check we have recieved all the packets and no duplicates.
	
	-- Hit process
	type HitStateType is (
		hit_idle,
		hit_write_record_1,
		hit_write_record_2,
		hit_write_record_3,
		hit_write_record_4,
		hit_write
	);
	signal hit_state : HitStateType;
	signal hit_count : unsigned (15 downto 0);
	signal hit_buff_addr : unsigned (ADDR_WIDTH downto 0);
	signal hit_buff_data : std_logic_vector (7 downto 0);
	signal hit_wants_bus : std_logic;
	signal hit_previous : std_logic;
	signal hit_record : std_logic_vector (31 downto 0); -- hit record to write, contains timer value on rising edge of hit.
	
	-- NMEA process (i.e. serial GPS messages coming in)
	type NMEAStateType is (
		nmea_idle,
		nmea_msg_type,
		nmea_msg_type_2,
		nmea_msg_type_3,
		nmea_msg_type_gps,
		nmea_msg_type_gps_2,
		nmea_msg_type_gps_3,
		nmea_msg_type_timing,
		nmea_gprmc,
		nmea_gprmc_write,
		nmea_gprmc_load,
		nmea_done
	);
	signal nmea_state : NMEAStateType;
	signal nmea_buff_addr : unsigned (ADDR_WIDTH downto 0);
	signal nmea_buff_data : std_logic_vector (7 downto 0);
	signal nmea_wants_bus : std_logic;
	signal nmeaMsgAddr_internal : unsigned (7 downto 0);
begin

	-- Memory bus arbiter
	bus_arbiter : process (
		pps_wants_bus,
		hit_wants_bus,
		nmea_wants_bus,
		pps_buff_data,
		pps_buff_addr,
		hit_buff_data,
		hit_buff_addr,
		nmea_buff_data,
		nmea_buff_addr
	)
	begin
		if (pps_wants_bus = '1') then
			-- pps has highest priority
			buff_write_enable <= '1';
			buff_data <= pps_buff_data;
			buff_addr <= pps_buff_addr;
		elsif (hit_wants_bus = '1') then
			buff_write_enable <= '1';
			buff_data <= hit_buff_data;
			buff_addr <= hit_buff_addr;	
		elsif (nmea_wants_bus = '1') then
			buff_write_enable <= '1';
			-- NMEA messages have lowest
			buff_data <= nmea_buff_data;
			buff_addr <= nmea_buff_addr;
		else
			buff_addr <= (others => '0');
			buff_data <= X"42";
			buff_write_enable <= '0';
		end if; 
	end process bus_arbiter;
	
-- Init ram from ucf file:	
--        INST "gt_?_I_gtstatus_decoder_I_ram" INIT_00 = "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff" ;
--        INST "gt_?_I_gtstatus_decoder_I_ram" INIT_01 = "eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee" ;
--        INST "gt_?_I_gtstatus_decoder_I_ram" INIT_02 = "dddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddd" ;
--        INST "gt_?_I_gtstatus_decoder_I_ram" INIT_03 = "cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc" ;
--			

	-- Packet buffer memory
   packet_buffer_process : process (clk, rst, buff_page, buff_write_enable)
	begin
		if (clk'event and clk='1') then
			-- Write into memory internally
			-- Read out externally from the opposite page to that being written to.
			-- NOTE: Must slice the right width out of buff_addr (9=1024, 11=4096).
			if (buff_write_enable = '1') then
				packet_buffer(to_integer(  ((not buff_page) & buff_addr(10 downto 0))  )) <= buff_data;
			end if;
			data <= packet_buffer(to_integer(unsigned(  (buff_page & addr(10 downto 0))  )));
		end if;
	end process packet_buffer_process;
		
	-- Pulse per second driven process, finalises packet
	-- with final timer count and packet number and then
	-- swaps buffers to output to PC.
	pps_process: process (pps, clk, rst)
	begin
		if (rst = '1') then
			timer <= (others => '0');
			timer_final <= (others => '0');
			packet_number <= (others => '0');
			newData <= '0';
			buff_page <= '0';
			pps_state <= pps_idle;
			pps_wants_bus <= '0';
			pps_buff_addr <= (others => '0');
			pps_buff_data <= (others => '0');			
		elsif (clk'event and clk = '1') then
			timer <= timer + 1;
			pps_previous <= pps;
			case pps_state is
			
				when pps_idle =>
					newData <= '0';
					pps_previous <= pps;
					if (pps = '1' and pps_previous = '0') then -- pps rising edge
						pps_wants_bus <= '1'; -- Grab bus
						timer_final <= std_logic_vector(timer); -- Save final timer value on pps edge.
						timer <= (others => '0'); -- rst timer immediately.
						pps_state <= pps_write_final_count_1;
					else
						pps_state <= pps_idle;
					end if;
				
				-- Write the final timer count value into the buffer.
				when pps_write_final_count_1 =>
					pps_buff_addr <= PACKET_FINAL_COUNT_ADDR;
					pps_buff_data <= "0000" & timer_final(27 downto 24);
					pps_state <= pps_write_final_count_2;
				when pps_write_final_count_2 =>
					pps_buff_addr <= PACKET_FINAL_COUNT_ADDR+1;
					pps_buff_data <= timer_final(23 downto 16);
					pps_state <= pps_write_final_count_3;
				when pps_write_final_count_3 =>
					pps_buff_addr <= PACKET_FINAL_COUNT_ADDR+2;
					pps_buff_data <= timer_final(15 downto 8);
					pps_state <= pps_write_final_count_4;
				when pps_write_final_count_4 =>
					pps_buff_addr <= PACKET_FINAL_COUNT_ADDR+3;
					pps_buff_data <= timer_final(7 downto 0);
					pps_state <= pps_write_packet_num_1;
					
				-- Write the packet number into the buffer.
				when pps_write_packet_num_1 =>
					pps_buff_addr <= PACKET_PACKET_NUM_ADDR;
					pps_buff_data <= std_logic_vector(packet_number(15 downto 8)); -- High byte
					pps_state <= pps_write_packet_num_2;
				when pps_write_packet_num_2 =>
					pps_buff_addr <= PACKET_PACKET_NUM_ADDR+1;
					pps_buff_data <= std_logic_vector(packet_number(7 downto 0)); -- Low byte
					pps_state <= pps_write_hit_count_1;
					
				-- Write the hit count into the buffer.
				when pps_write_hit_count_1 =>
					pps_buff_addr <= PACKET_HIT_COUNT_ADDR;
					pps_buff_data <= std_logic_vector(hit_count(15 downto 8)); -- High byte
					pps_state <= pps_write_hit_count_2;
				when pps_write_hit_count_2 =>
					pps_buff_addr <= PACKET_HIT_COUNT_ADDR+1;
					pps_buff_data <= std_logic_vector(hit_count(7 downto 0)); -- Low byte
					pps_state <= pps_newData;
					
				when pps_newData =>
					-- Swap buffers and assert new packet.
					newData <= '1';
					buff_page <= not buff_page;
					packet_number <= packet_number + 1;
					-- Done!
					pps_wants_bus <= '0'; -- Release bus
					pps_state <= pps_newData_2;
				when pps_newData_2 =>
					-- Wait one more cycle for newData to be registered.
					pps_state <= pps_idle;
					
			end case;	
		end if;
	end process pps_process;
	
	nmeaMsgAddr <= std_logic_vector(nmeaMsgAddr_internal);
	nmea_process: process (nmeaMsgValid, clk, rst, pps_state)
	begin
		-- Also rst if we start a new packet so we don't carry on writing into
		-- the buffer after rst
		if (rst = '1' or pps_state = pps_newData) then
			nmea_wants_bus <= '0';
			nmea_buff_addr <= (others => '0');
			nmea_buff_data <= (others => '0');
			nmeaMsgAddr_internal <= (others => '0');
		elsif (clk'event and clk = '1') then
			case nmea_state is

				when nmea_idle =>
					-- Wait for nmeaMsgValid
					if (nmeaMsgValid = '1') then
						nmea_state <= nmea_msg_type;
					else
						nmea_state <= nmea_idle;
					end if;
					
				when nmea_msg_type =>
					-- Identify meg type and branch
					nmeaMsgAddr_internal <= X"00"; -- Ask for msg index 0 and wait a clock cycle
					nmea_state <= nmea_msg_type_2;
				when nmea_msg_type_2 =>
					-- waiting...
					nmea_state <= nmea_msg_type_3;
				when nmea_msg_type_3 =>
					-- Check 0th char is a 'G' indicating a GPS message.
					-- If so goto msg_type_gps, else msg_type_timing.
					-- ASCII 'G' = 0x47
					if (nmeaMsgData = X"47") then
						nmea_state <= nmea_msg_type_gps;
					else
						nmea_state <= nmea_msg_type_timing;
					end if;
					
				when nmea_msg_type_gps =>
					-- Check GPS meg type, look for RMC
					nmeaMsgAddr_internal <= X"02";
					nmea_state <= nmea_msg_type_gps_2;
				when nmea_msg_type_gps_2 =>
					nmea_state <= nmea_msg_type_gps_3;
				when nmea_msg_type_gps_3 =>
					-- If 2nd char is 'R' assume GPRMC
					-- ASCII 'R' = 0x52
					if (nmeaMsgData = X"52") then
						nmea_state <= nmea_gprmc;
					else
						-- else ignore
						nmea_state <= nmea_done;
					end if;
					
				when nmea_msg_type_timing =>
					-- Could check that we have the right msg type here but
					-- for now lets just assume its all ok.
					-- Extract mode (surveying?) and timing offset and load into
					-- packet
					nmea_state <= nmea_done; -- FIXME: ignoring for now.
					
				when nmea_gprmc =>
					-- For now just extract the time of day and store in packet.
					-- Time is in bytes 6-11 as an ASCII string, easier to process on the
					-- host side so just copy this into the packet
					nmeaMsgAddr_internal <= X"06";
					nmea_buff_addr <= PACKET_TIME_ADDR-1;
					nmea_state <= nmea_gprmc_load;
				when nmea_gprmc_write =>
					if (nmeaMsgAddr_internal <= X"0B") then
						nmeaMsgAddr_internal <= nmeaMsgAddr_internal + 1;
						nmea_buff_addr <= nmea_buff_addr + 1;
						nmea_buff_data <= nmeaMsgData;
						nmea_wants_bus <= '1'; -- Grab bus
						nmea_state <= nmea_gprmc_load;
					else
						nmea_state <= nmea_done;
						nmea_wants_bus <= '0'; -- Release bus
					end if;
				when nmea_gprmc_load =>
					-- Wait one cycle to load from msg memory.
					nmea_state <= nmea_gprmc_write;
					
				when nmea_done =>
					-- Wait until msg not valid and start again.
					if (nmeaMsgValid = '0') then
						nmea_state <= nmea_idle;
					else
						nmea_state <= nmea_done;
					end if;
				
			end case;
		end if;
	end process nmea_process;

	hit_process: process (hit, clk, rst, pps_state)
		variable hit_any : STD_LOGIC;
	begin
		hit_any := hit(0) or hit(1) or hit(2) or hit(3);
		-- Also rst if we start a new packet so we don't carry on writing into
		-- the buffer after rst
		if (rst = '1' or pps_state = pps_newData) then
			hit_state <= hit_idle;
			hit_count <= (others => '0');
			hit_buff_addr <= (others => '0');
			hit_buff_data <= (others => '0');
			hit_wants_bus <= '0';
		elsif (clk'event and clk = '1') then
			hit_previous <= hit_any;
			if (pps_state = pps_newData) then
				hit_count <= (others => '0');
			end if;
			case hit_state is
			
				when hit_idle =>
					if (hit_any = '1' and hit_previous = '0') then -- Rising edge
						hit_wants_bus <= '1'; -- Grab bus
						hit_buff_addr <= PACKET_HITS_START_ADDR + (hit_count sll 2);
						-- For now lets just use hit as the hit channel number, this
						-- will allow us to see collisions directly.
						hit_record <= X"DEADBEEF";--hit & std_logic_vector(timer);
						hit_state <= hit_write_record_1;
					else
						hit_wants_bus <= '0'; -- Release bus
						hit_state <= hit_idle;
					end if;

				-- Write the hit record into the buffer.
				when hit_write_record_1 =>
					hit_buff_data <= hit_record(31 downto 24); --hit_channel & hit_record(27 downto 24);
					hit_state <= hit_write_record_2;
				when hit_write_record_2 =>
					hit_buff_addr <= hit_buff_addr + 1;
					hit_buff_data <= hit_record(23 downto 16);
					hit_state <= hit_write_record_3;
				when hit_write_record_3 =>
					hit_buff_addr <= hit_buff_addr + 1;
					hit_buff_data <= hit_record(15 downto 8);
					hit_state <= hit_write_record_4;
				when hit_write_record_4 =>
					hit_buff_addr <= hit_buff_addr + 1;
					hit_buff_data <= hit_record(7 downto 0);
					hit_state <= hit_write;

				when hit_write =>
					-- Increment the hit count after writing in case we get
					-- interrupted by a buffer swap
					hit_count <= hit_count + 1;
					-- Wait cycle to write
					hit_state <= hit_idle;
			
			end case;
		end if;
	end process hit_process;

end Behavioral;








--	test_process: process (clk, rst)
--	begin
--		if (rst = '1') then
--			buff_page <= '0';
--			buff_addr <= X"0000";
--			buff_data <= X"00";
--		elsif (clk'event and clk = '1') then
--			buff_addr <= buff_addr + 1;
--			
--			if (buff_page = '0') then
--				buff_data <= std_logic_vector(buff_addr+1);
--			else
--				buff_data <= X"00";
--			end if;
--			
--			if (buff_addr > GPSTime_buff_size) then
--				buff_page <= not buff_page;
--				buff_addr <= X"0000";
--				buff_data <= (others => '0');
--			end if;
--		end if;
--	end process test_process;