----------------------------------------------------------------------------------
-- CROCS - Globals
-- Various system wide constants and defines.
--
-- Fergus Noble 2011
----------------------------------------------------------------------------------

LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.ALL;

package Globals is

	-- Global constants
	constant clk_freq : integer := 50000000; -- 50 MHz xtal
	constant clk_period : time := 1 sec / clk_freq;
	constant clk_half_period : time := clk_period / 2;
	
	-- NMEA decoder related
	constant NMEA_max_len : integer := 80;
	-- (NMEA max length 82 inc. $ and <CR><LF>, i.e. 79 bytes of message body then add 1 byte null termination)
	type NMEAMsgType is array (0 to NMEA_max_len) of std_logic_vector (7 downto 0);
	
	-- GPSTime related
	constant GPSTime_buff_size : integer := 1024;--4096;
	type GPSTimePacketType is array (0 to 2*GPSTime_buff_size) of std_logic_vector (7 downto 0); -- (MSB selects page)
	
	constant PACKET_PACKET_NUM_ADDR : unsigned := X"0000"; -- 2 bytes
	constant PACKET_FINAL_COUNT_ADDR : unsigned := X"0002"; -- 4 bytes
	constant PACKET_HIT_COUNT_ADDR : unsigned := X"0006"; -- 2 bytes
	constant PACKET_TIME_ADDR : unsigned := X"0008"; -- 6 bytes	
	constant PACKET_HITS_START_ADDR : unsigned := X"0020"; -- hit_count * 4 bytes 
	
	-- SPI / USB related
	constant SPI_NULL_OPCODE : unsigned (7 downto 0) := X"00";
	constant SPI_DATA_READOUT_OPCODE : unsigned (7 downto 0) := X"42";
	
	-- converts a hex nibble in ASCII representation to an unsigned
	function asciihex2unsigned_nibble(hex : std_logic_vector) return unsigned;
	
	-- converts a hex byte in ASCII representation to an unsigned
	function asciihex2unsigned(hex_high, hex_low : std_logic_vector) return unsigned;
	
end Globals;

package body Globals is
	
	function asciihex2unsigned_nibble(hex : std_logic_vector) return unsigned is
	begin
		-- '0' = 0x30, 9 = 0x39 
		if (unsigned(hex) - X"30" < 10) then
			-- nibble is 0-9
			return unsigned(hex) - X"30";
		-- 'A' = 0x41, 'F' = 0x46
		elsif (unsigned(hex) - X"41" < 6) then
			-- nibble is A-F
			return unsigned(hex) - X"41" + 10;
		else
			-- invalid character for checksum byte (wanted 0-9, A-F)
			-- TODO: find a way of handling this error
			return X"00";
		end if;
	end asciihex2unsigned_nibble;
	
	function asciihex2unsigned(hex_high, hex_low : std_logic_vector) return unsigned is
	begin
		return asciihex2unsigned_nibble(hex_low) + (asciihex2unsigned_nibble(hex_high) SLL 4);
	end asciihex2unsigned;

end Globals;
