----------------------------------------------------------------------------------
-- CROCS - NMEA138Decode
-- Finds the start of NMEA-138 packets and verifies their checksums.
--
-- Fergus Noble 2011
----------------------------------------------------------------------------------

LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.ALL;

LIBRARY work;
USE work.Globals.ALL;

entity NMEA183Decode is	
    Port ( data : in  STD_LOGIC_VECTOR (7 downto 0);
           dataReady : in  STD_LOGIC;
			  rst : in  STD_LOGIC;
			  clk : in STD_LOGIC;
           msgValid : out  STD_LOGIC;
			  msgData : out STD_LOGIC_VECTOR (7 downto 0);
			  msgAddr : in STD_LOGIC_VECTOR (7 downto 0)
			);
end NMEA183Decode;

architecture Behavioral of NMEA183Decode is
	type ReadMsgStateType is (idle, get_byte, checksum_1, checksum_2, packet_end_1, packet_end_2);
	signal read_msg_state : ReadMsgStateType;
	signal data_read : std_logic;	-- toggle to ensure we only read the new byte once
	
	signal reading_msg : std_logic;
	signal checksum, checksum_high : std_logic_vector (7 downto 0);
	
	signal msg_buffer : NMEAMsgType;
	signal msg_buffer_index : integer range 0 to NMEA_max_len;
	signal msg_buffer_data : std_logic_vector (7 downto 0);
begin
	-- Reads NMEA message into the msg buffer
	read_msg_process: process (clk, rst)
	begin
		if (rst = '1') then
			read_msg_state <= idle;
			data_read <= '0';
			
			reading_msg <= '0';
			checksum <= (others => '0');
			checksum_high <= (others => '0');
			
			msgValid <= '0';
			msg_buffer_index <= 0;
			msg_buffer_data <= (others => '0');

		elsif (clk'event and clk = '1') then
			if (dataReady = '0') then
				data_read <= '0';
			elsif (data_read = '0') then	-- only run the state machine once per new byte
				data_read <= '1';
				case read_msg_state is
				
					when idle =>
						reading_msg <= '0';
						-- Wait for a '$' marking the beginning of a packet ('$' = 0x24)
						if (unsigned(data) = X"24") then
							read_msg_state <= get_byte;
							msg_buffer_index <= NMEA_max_len;
							msgValid <= '0';
							reading_msg <= '1';
							checksum <= (others => '0');
						else
							read_msg_state <= idle;
						end if;
						
					when get_byte =>
						-- Wait for a '*' (0x2A)
						if (unsigned(data) = X"2A") then
							-- Null terminate the string
							msg_buffer_data <= X"00";
							msg_buffer_index <= msg_buffer_index + 1;
							
							read_msg_state <= checksum_1;
						else
							-- Read in another byte of message
							checksum <= checksum XOR data;
							
							msg_buffer_data <= data;
							
							-- Wrap buffer, also gives us a starting index of zero.
							-- Bit of a hack but does the job.
							if (msg_buffer_index = NMEA_max_len) then
								msg_buffer_index <= 0;
							else
								msg_buffer_index <= msg_buffer_index + 1;
							end if;
							
							read_msg_state <= get_byte;
						end if;
						
					when checksum_1 =>
						-- Checksum high nibble now in data
						checksum_high <= data;
						read_msg_state <= checksum_2;
						
					when checksum_2 =>
						-- Checksum low nibble now in data
						if (unsigned(checksum) = asciihex2unsigned(checksum_high, data)) then
							-- Checksum OK
							read_msg_state <= packet_end_1;
						else
							-- error, checksum failed
							read_msg_state <= idle;
						end if;
						
					when packet_end_1 =>
						-- Look for a <CR> (0x0D)
						if (unsigned(data) = X"0D") then
							read_msg_state <= packet_end_2;
						else
							-- Error, should have seen an <CR>
							read_msg_state <= idle;
						end if;
						
					when packet_end_2 =>
						-- Found a <CR>, if we now have an <LF> (0x0A) its the end of the packet
						if (unsigned(data) = X"0A") then
							reading_msg <= '0';
							msgValid <= '1';
							read_msg_state <= idle;
						else
							-- Error, should have seen an <LF>
							read_msg_state <= idle;
						end if;
						
				end case;
			end if;
		end if;
	end process read_msg_process;	

	-- memory example
   msg_buffer_process : process (clk, rst)
	begin
		if (clk'event and clk='1') then
			msg_buffer(msg_buffer_index) <= msg_buffer_data;
			--if (unsigned(msgAddr) <= NMEA_max_len) then
				msgData <= msg_buffer(to_integer(unsigned(msgAddr)));
			--end if;
		end if;
	end process msg_buffer_process;

end Behavioral;

