----------------------------------------------------------------------------------
-- CROCS - NMEA138Decode
-- Finds the start of NMEA-138 packets and verifies their checksums.
--
-- Fergus Noble 2011
----------------------------------------------------------------------------------

LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.ALL;

entity NMEA183Decode is
    Port ( data : in  STD_LOGIC_VECTOR (7 downto 0);
           data_ready : in  STD_LOGIC;
			  reset : in  STD_LOGIC;
           packet_start : out  STD_LOGIC;
           checksum_valid : out  STD_LOGIC);
end NMEA183Decode;

architecture Behavioral of NMEA183Decode is
	type StateType is (idle, get_byte, chksum_byte1, chksum_byte2, packet_check);
	signal state : StateType;
	
	signal data_foo : unsigned (7 downto 0);
	signal chksum : unsigned (7 downto 0);
	signal chksum_in : unsigned (7 downto 0);
	signal count : unsigned (7 downto 0);

	type msg_type_type is array (5 downto 0) of std_logic_vector (7 downto 0);
	signal msg_type : msg_type_type; 
begin

--	process (reset)
--	begin
--		state <= idle;
--	end process;

	process (data_ready)
	begin
		if (data_ready'event and data_ready = '1') then
			data_foo <= unsigned(data);
			case state is
				
				when idle =>
					packet_start <= '0';
					checksum_valid <= '0';
					-- Wait for a '$' marking the beginning of a packet ('$' = 0x24)
					if (data_foo = X"24") then
						chksum <= X"00";	-- reset the checksum
						state <= get_byte;
						packet_start <= '1';
					else
						state <= idle;
					end if;
					
				--when get_type =>
					-- Read in the next 5 characters containing the message type
					
					
				when get_byte =>
					-- Wait for a '*' (0x2A) and XOR together bytes to calculate checksum
					packet_start <= '0';
					if (data_foo = X"2A") then
						state <= chksum_byte1;
					else
						chksum <= chksum XOR data_foo;
						state <= get_byte;
					end if;
					
				when chksum_byte1 =>
					-- Check the high nibble of the checksum against the recieved hex nibble.
					-- '0' = 0x30, 9 = 0x39 
					if (data_foo - X"30" < 10) then
						-- first nibble is 0-9
						chksum_in <= (data_foo - X"30") SLL 4;
						state <= chksum_byte2;
					-- 'A' = 0x41, 'F' = 0x46
					elsif (data_foo - X"41" < 6) then
						-- first nibble is A-F
						chksum_in <= (data_foo - X"41" + 10) SLL 4;
						state <= chksum_byte2;
					else
						-- invalid character for checksum byte (wanted 0-9, A-F)
						state <= idle;
					end if;
					
				when chksum_byte2 =>
					-- Read in low nibble and add it to our chksum_in.
					-- '0' = 0x30, 9 = 0x39 
					if (data_foo - X"30" < 10) then
						-- second nibble is 0-9
						chksum_in <= chksum_in + (data_foo - X"30");
						state <= packet_check;
					-- 'A' = 0x41, 'F' = 0x46
					elsif (data_foo - X"41" < 6) then
						-- second nibble is A-F
						chksum_in <= chksum_in + (data_foo - X"41" + 10);
						state <= packet_check;
					else
						-- invalid character for checksum byte (wanted 0-9, A-F)
						state <= idle;
					end if;
					
				when packet_check =>
					-- Now see if the checksum received matches our calculated one.
					if (chksum = chksum_in) then
						-- packet is ok!
						checksum_valid <= '1';
						state <= idle;
					else
						-- fail
						state <= idle;
					end if;
					
			end case;
		end if;
	end process;		
end Behavioral;

