----------------------------------------------------------------------------------
-- CROCS - SPIDataReadout
-- Allows readout of a data buffer over SPI.
--
-- Fergus Noble 2011, modified from Cosmix_lib.SPI_CTR.Mk2 M. Goodrick 2005
----------------------------------------------------------------------------------

LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.ALL;

LIBRARY work;
USE work.Globals.ALL;

entity SPIDataReadout is
	 Generic ( BUFF_SIZE : unsigned (15 downto 0) := X"0800"; -- 2048 byte buffer.
				  MY_OPCODE : unsigned (7 downto 0) := X"16"		 -- Opcode to respond to.
				);
    Port ( reset : in  STD_LOGIC;
           clk : in  STD_LOGIC;
			  
           spck : in  STD_LOGIC;
           spdi : in  STD_LOGIC;
           spcs : in  STD_LOGIC;
           spdo : out  STD_LOGIC;
			  
			  new_packet : in STD_LOGIC;
			  addr : out STD_LOGIC_VECTOR (15 downto 0);
			  data : in STD_LOGIC_VECTOR (7 downto 0)
			);
end SPIDataReadout; 

architecture Behavioral of SPIDataReadout is
   type StateType is (
		idle,
		check_opcode,
		get_arg,
		branch_arg,
		-- branch to one of these states depending on arg:
		
		-- resets the readout, i.e. puts pointer into buffer back to the beginning
		do_reset,
		
		-- send a byte 0x00 or 0x01 depending on if there is a new packet waiting
		do_packet_ready,
		do_packet_ready_send,
		
		-- send the size of the buffer as a 16-bit word
		do_buff_size,
		do_buff_size_msb,
		do_buff_size_lsb,
		
		-- send the next nreads bytes of the buffer
		do_read,
		do_read_save_nreads,
		do_read_send,
		do_read_send_wait_1,
		do_read_send_wait_2,
		
		-- helper states
		get_byte,
		send_byte,
		-- these states just wait for reset on CS
		finished,
		not_me
	);
   signal state, state_after : StateType;
	
	signal serial_reg : std_logic_vector (7 downto 0);
   signal bit_count : integer range 0 to 7;
	signal opcode, arg, nreads : unsigned (7 downto 0);
	signal spck_l, spck_ll : std_logic; -- Previous spck states (for the last and second last clock cycles).
	
	signal packet_ready : std_logic;
	signal new_packet_previous : std_logic;
	signal buffer_ptr : unsigned (15 downto 0);
begin

	state_machine: process (clk, reset, spcs, spck, new_packet, buffer_ptr)
	begin
		addr <= std_logic_vector(buffer_ptr);
		if (clk'event and clk = '1') then
		
			new_packet_previous <= new_packet;
			
			if (reset = '1' or spcs = '1') then
				-- Reset
				state <= idle;
				state_after <= idle;
				serial_reg <= (others => '0');
				bit_count <= 0;
				opcode <= X"00";
				spck_l <= spck;
				spck_ll <= spck;
				spdo <= 'Z';
				
				-- TODO: with this here we drop the occasional packet, 
				-- if we put it outside the reset block then if a new_packet edge
				-- comes when we are sending the data then it gets very confused.
				-- maybe needs to be outside with a way to defer until after the current operation is finished
				if (new_packet = '1' and new_packet_previous = '0') then
					-- If we have a rising edge on the new_packet input then
					-- flag that we have a packet ready.
					packet_ready <= '1';
					buffer_ptr <= X"0000";
				end if;

				if (reset = '1') then
					-- Hard reset from internal reset source, not SPI bus reset (i.e. CS high)
					-- (these things need to persist across SPI frames)
					packet_ready <= '0';
					buffer_ptr <= X"0000";
				end if;
			else
				-- Save the previous spck states
				spck_l <= spck;
				spck_ll <= spck_l;
				
				case state is
				
					when idle =>
						spdo <= 'Z'; -- may not be for us
						if (spck_ll = '0' and spck_l = '1') then  -- spck rising edge
							serial_reg <= serial_reg(6 downto 0) & spdi; -- shift in next bit
							if(bit_count = 7) then
								bit_count <= 0;
								opcode <= unsigned(serial_reg(6 downto 0) & spdi);
								state <= check_opcode;
							else
								bit_count <= bit_count + 1;
							end if;                
						end if;
						
					when check_opcode =>
						if (opcode(6 downto 0) = MY_OPCODE(6 downto 0)) then -- Check 7 LSBs, 8th bit indicates write. 
							state <= get_arg;
						else
							state <= not_me;
						end if;					
						
					when get_arg =>
						-- Read in the argument byte
						bit_count <= 0;
						state <= get_byte;
						state_after <= branch_arg;
					when branch_arg =>
						-- Branch on arg.
						arg <= unsigned(serial_reg);
						case serial_reg is
							when X"00" =>
								state <= do_reset;
							when X"01" =>
								state <= do_packet_ready;
							when X"02" =>
								state <= do_buff_size;
							when X"03" =>
								state <= do_read;
							when others =>
								-- if we don't understand the arg then just ignore the packet.
								state <= not_me;
						end case;
						
					when do_reset =>
						buffer_ptr <= X"0000";
						state <= finished;
						
					when do_packet_ready =>
						-- discard nreads
						bit_count <= 0;
						state <= get_byte;
						state_after <= do_packet_ready_send;
					when do_packet_ready_send =>
						-- send packet ready status
						serial_reg <= (others => packet_ready);
						state_after <= finished;
						bit_count <= 0;
						state <= send_byte;
						
					when do_buff_size =>
						-- discard nreads
						bit_count <= 0;
						state <= get_byte;
						state_after <= do_buff_size_lsb;					
					when do_buff_size_lsb =>
						-- send the size of the buffer.
						-- send MSB first
						serial_reg <= std_logic_vector(BUFF_SIZE(7 downto 0));
						state_after <= do_buff_size_msb;
						bit_count <= 0;
						state <= send_byte;
					when do_buff_size_msb =>
						-- now send LSB
						serial_reg <= std_logic_vector(BUFF_SIZE(15 downto 8));
						state_after <= finished;
						bit_count <= 0;
						state <= send_byte;					
						
					when do_read =>
						-- get nreads first
						bit_count <= 0;
						state <= get_byte;
						state_after <= do_read_save_nreads;	
					when do_read_save_nreads =>
						nreads <= unsigned(serial_reg);
						state <= do_read_send;
					when do_read_send =>
						packet_ready <= '0'; -- started reading in the packet so it is no longer new.
						if (nreads > 0) then
							if (buffer_ptr < BUFF_SIZE) then
								bit_count <= 0;
								state <= send_byte;
								nreads <= nreads - 1;
								buffer_ptr <= buffer_ptr + 1;
								serial_reg <= data;
								state_after <= do_read_send_wait_1;
							else
								state <= finished;
							end if;
						else
							state <= finished;
						end if;
					when do_read_send_wait_1 =>
						state <= do_read_send_wait_2;
					when do_read_send_wait_2 =>
						state <= do_read_send;
		

					when get_byte =>
						-- Get one byte, put it in serial_reg and then jump to state_after
						spdo <= '0';
						if (spck_ll = '0' and spck_l = '1') then  -- spck rising edge
							serial_reg <= serial_reg(6 downto 0) & spdi; -- shift in next bit
							if(bit_count = 7) then
								bit_count <= 0;
								state <= state_after;
							else
								bit_count <= bit_count + 1;
							end if;                
						end if;
						
					when send_byte =>
						-- Send byte in serial_reg and then jump to state_after.
						if (spck_ll = '1' and spck_l = '0') then  -- spck falling edge
							spdo <= serial_reg(7);
							serial_reg <= serial_reg(6 downto 0) & '0'; -- shift serial_reg along
							if (bit_count = 7) then
								state <= state_after;
							else
								bit_count <= bit_count + 1;
							end if;
						end if;
						
					when finished =>
						-- Now just wait for reset on CS low.
						state <= finished;
						
					when not_me =>
						-- Packet isn't ment for us, wait until reset when CS goes low.
						spdo <= 'Z';
				
				end case;
			end if;
		end if;
	end process state_machine;
end Behavioral;

