----------------------------------------------------------------------------------
-- CROCS - Globals
-- Various helper functions for writing testbenches.
--
-- Fergus Noble 2011
----------------------------------------------------------------------------------

LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.ALL;

LIBRARY work;
USE work.Globals.ALL;

package SimHelper is
	-- Utility functions

	-- Convert a character type to a std_logic_vector
	function char2vector(chr: character) return std_logic_vector;
	-- Creates an async serial character stimulus for simulation
	procedure sim_serial_char(chr: in character; baud_rate: in integer; signal serial_tx : out std_logic);
	-- Creates an async serial string stimulus for simulation
	procedure sim_serial_string(data: in string; baud_rate: in integer; signal serial_tx : out std_logic);
	
	procedure sim_spi_start(clk_freq : integer; signal clk, cs, mosi : out std_logic);
	procedure sim_spi_end(clk_freq : integer; signal clk, cs, mosi : out std_logic);
	procedure sim_spi(data: in unsigned(7 downto 0); clk_freq: in integer; signal clk, cs, mosi : out std_logic);
	procedure sim_spi_string(data: in string; clk_freq: in integer; signal clk, cs, mosi : out std_logic);

	procedure sim_read_msg(
		opcode, arg : unsigned (7 downto 0);
		nreads : integer;
		clk_freq: integer;
		signal clk, cs, mosi : out std_logic
	);
	procedure sim_write_msg(
		opcode, arg, byte1, byte2 : unsigned (7 downto 0);
		clk_freq: integer;
		signal clk, cs, mosi : out std_logic
	);
	
end SimHelper;

package body SimHelper is

	function char2vector(chr: character) return std_logic_vector is
	begin
		return std_logic_vector(to_unsigned(character'pos(chr),8));
	end char2vector;
	
	procedure sim_serial_char(
		chr : in character;					 -- character to send
		baud_rate : integer;					 -- baud rate
		signal serial_tx : out std_logic  -- serial output signal
	) is
		variable chr_vector : std_logic_vector(7 downto 0);
		constant bit_period : time := 1 sec / baud_rate;
	begin
		chr_vector := char2vector(chr);
		-- start bit
		serial_tx <= '0';
		wait for bit_period;
		
		-- data bits, MSB first
		for i in chr_vector'reverse_range loop
			serial_tx <= chr_vector(i);
			wait for bit_period;
		end loop;
		
		-- stop bit
		serial_tx <= '1';
		wait for bit_period;
	end; -- procedure sim_serial_char
	
	procedure sim_serial_string(
		data : in string;						 -- string to send
		baud_rate : integer;					 -- baud rate
		signal serial_tx : out std_logic  -- serial output signal
	) is
		constant bit_period : time := 1 sec / baud_rate;
	begin
		-- ensure we have a bit of a pause before the string
		serial_tx <= '1';
		wait for bit_period*10;
		
		for i in data'range loop
			sim_serial_char(data(i), baud_rate, serial_tx);
		end loop;
	end; -- procedure sim_serial_string
	
	procedure sim_spi_start(
		clk_freq : integer;
		signal clk, cs, mosi : out std_logic
	) is
		constant bit_period : time := 1 sec / clk_freq;
	begin	
		-- ensure we have a bit of a pause before the string
		mosi <= '1';
		clk <= '1';
		cs <= '1';
		wait for bit_period*10;
		
		-- assert CS
		cs <= '0';
		wait for 50ns;
	end;
	
	procedure sim_spi_end(
		clk_freq : integer;
		signal clk, cs, mosi : out std_logic
	) is
		constant bit_period : time := 1 sec / clk_freq;
	begin	
		-- de-assert CS
		wait for 50ns;
		cs <= '1';
	end;

	procedure sim_spi(
		data : in unsigned (7 downto 0);	 -- byte to send
		clk_freq : integer;
		signal clk, cs, mosi : out std_logic
	) is
		constant bit_period : time := 1 sec / clk_freq;
		variable chr_vector : std_logic_vector(7 downto 0);
	begin

		for j in data'range loop
			mosi <= data(j);
			clk <= '0';
			wait for bit_period/2;
			clk <= '1';
			wait for bit_period/2;
		end loop;	
		wait for 2*bit_period;

	end; -- procedure sim_spi
	
	procedure sim_spi_string(
		data : in string;						 -- string to send
		clk_freq : integer;
		signal clk, cs, mosi : out std_logic
	) is
		constant bit_period : time := 1 sec / clk_freq;
		variable chr_vector : std_logic_vector(7 downto 0);
	begin

		for i in data'range loop
			chr_vector := char2vector(data(i));
			for j in chr_vector'reverse_range loop
				mosi <= chr_vector(j);
				clk <= '0';
				wait for bit_period/2;
				clk <= '1';
				wait for bit_period/2;
			end loop;
			wait for 2*bit_period;
		end loop;
	end; -- procedure sim_spi_string

	procedure sim_read_msg(
		opcode, arg : unsigned (7 downto 0);
		nreads : integer;
		clk_freq: integer;
		signal clk, cs, mosi : out std_logic
	) is
	begin
		wait for 200ns;
		sim_spi_start(clk_freq, clk, cs, mosi);
		
		-- header
		sim_spi(opcode, clk_freq, clk, cs, mosi);
		sim_spi(arg, clk_freq, clk, cs, mosi);
		sim_spi(to_unsigned(nreads,8), clk_freq, clk, cs, mosi);
		
		-- nreads * chuff while reading
		for i in nreads downto 0 loop
			sim_spi(X"00", clk_freq, clk, cs, mosi);
		end loop;
		
		sim_spi_end(clk_freq, clk, cs, mosi);
		wait for 200ns;
	end procedure;
	
	procedure sim_write_msg(
		opcode, arg, byte1, byte2 : unsigned (7 downto 0);
		clk_freq: in integer;
		signal clk, cs, mosi : out std_logic
	) is
	begin
		wait for 200ns;
		sim_spi_start(clk_freq, clk, cs, mosi);
		
		-- header
		sim_spi(opcode, clk_freq, clk, cs, mosi);
		sim_spi(arg, clk_freq, clk, cs, mosi);
		
		-- two bytes to write
		sim_spi(byte1, clk_freq, clk, cs, mosi);
		sim_spi(byte2, clk_freq, clk, cs, mosi);
		
		sim_spi_end(clk_freq, clk, cs, mosi);
		wait for 200ns;
	end procedure;

end SimHelper;
