/*
Cosmix data logger control panel

Steve Wotton, 2008
Cavendish Lab (HEP)
*/

package mjg.instrument;

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.Graphics2D.*;
import java.net.*;
import java.applet.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.html.*;
import javax.swing.border.*;
import java.text.DateFormat;
import java.util.Date;
import java.util.prefs.Preferences;
import org.jfree.data.time.Minute;
import mjg.instrument.*;

public class DataLoggerControl extends JPanel implements ActionListener
{
  Preferences prefs;
  
  TimeSeriesChart rateChart;
  TimeSeriesChart pressureChart;

  JTextField fileNameTF;
  JTextField intervalTF;
  JProgressBar progressBar;
  JButton stopButton;
  JButton startButton;
  JButton pauseButton;
  JButton resumeButton;

  FileWriter dataStream;

  long loggingStartTime;
  int loggingInterval;

  public DataLoggerControl()
  {
    JButton button;

    prefs = Preferences.userNodeForPackage(DataLoggerControl.class).node("CosmixDataLoggerControl");

    setLayout( new GridBagLayout() );
    GridBagConstraints gbC = new GridBagConstraints();
    gbC.insets = new Insets(2,2,2,2);
    gbC.weightx = 1.0;

    JPanel filePanel = new JPanel();
    filePanel.setBorder(BorderFactory.createTitledBorder("Output file"));

    fileNameTF = new JTextField( prefs.get("fileName","cosmix.csv"), 50);
    fileNameTF.setEditable(true);
    fileNameTF.addActionListener( this );
    fileNameTF.setToolTipText( "Output file name" );
    filePanel.add( fileNameTF );

    button = new JButton("Browse...");
    button.setActionCommand("browse");
    button.addActionListener( this );
    //    gbC.gridwidth = GridBagConstraints.REMAINDER;
    filePanel.add( button, gbC );
    //    gbC.gridwidth = 1;

    add( filePanel, gbC );

    JPanel intervalPanel = new JPanel();
    intervalPanel.setBorder(BorderFactory.createTitledBorder("Log interval [sec]"));

    loggingInterval = prefs.getInt("interval",10);
    intervalTF = new JTextField( String.format("%1$d",loggingInterval), 10);
    intervalTF.setEditable(true);
    intervalTF.setActionCommand("interval");
    intervalTF.addActionListener( this );
    intervalTF.setToolTipText( "Time between logging data records in seconds" );
    intervalPanel.add( intervalTF );

    gbC.gridwidth = GridBagConstraints.REMAINDER;
    add( intervalPanel, gbC );
    gbC.gridwidth = 1;

    progressBar = new JProgressBar();
    gbC.gridwidth = GridBagConstraints.REMAINDER;
    gbC.fill = GridBagConstraints.HORIZONTAL;
    add( progressBar, gbC );
    gbC.gridwidth = 1;
    gbC.fill = GridBagConstraints.NONE;

    rateChart = new TimeSeriesChart("Rate [Hz]");
    gbC.gridwidth = GridBagConstraints.REMAINDER;
    add(rateChart.getChartPanel(), gbC);
    gbC.gridwidth = 1;

    pressureChart = new TimeSeriesChart("Pressure [mbar]");
    gbC.gridwidth = GridBagConstraints.REMAINDER;
    add(pressureChart.getChartPanel(), gbC);
    gbC.gridwidth = 1;

    JPanel buttonPanel = new JPanel();
    buttonPanel.setLayout(new GridLayout(1,4));

    startButton = new JButton( "Start" );
    startButton.addActionListener( this );
    startButton.setActionCommand( "start" );
    startButton.setEnabled(true);
    buttonPanel.add( startButton );

    stopButton = new JButton( "Stop" );
    stopButton.addActionListener( this );
    stopButton.setActionCommand( "stop" );
    stopButton.setEnabled(false);
    buttonPanel.add( stopButton );

    pauseButton = new JButton( "Pause" );
    pauseButton.addActionListener( this );
    pauseButton.setActionCommand( "pause" );
    pauseButton.setEnabled(false);
    buttonPanel.add( pauseButton );

    resumeButton = new JButton( "Resume" );
    resumeButton.addActionListener( this );
    resumeButton.setActionCommand( "resume" );
    resumeButton.setEnabled(false);
    buttonPanel.add( resumeButton );


    gbC.gridwidth = GridBagConstraints.REMAINDER;
    add( buttonPanel, gbC );
    gbC.gridwidth = 1;

    return;
  }

public void closeStream()
{

  try
  {

// Close and destroy the stream if already exists

    if ( dataStream != null ) { dataStream.close(); dataStream = null; }

  }
  catch ( IOException ioe )
  {
     System.err.println("Warning: " + ioe.getMessage() );
  }

  return;
}

public void openStream()
{

  String fileName = fileNameTF.getText();
  System.out.println("Opening " + fileName);

  try
  {

// Close and destroy the stream if already exists

    if ( dataStream != null ) { dataStream.close(); dataStream = null; }

// Open in append mode

    dataStream = new FileWriter(fileName);
    String hdr = new String("DD/MM/YYYY HH:MM:SS , time , count0 , count1 , count2 , count3 , count4 , count5 , adc0 , adc1 , adc2 , adc3\n");
    dataStream.write(hdr,0,hdr.length());

  }
  catch ( IOException ioe )
  {
     System.err.println("Warning: " + ioe.getMessage() );
  }

  return;
}

public void updatePreferences()
{
  String fileName = fileNameTF.getText();
  prefs.put("fileName", fileName);

  return;
}

  public void actionPerformed( ActionEvent e )
  {
    int ival;
    boolean flag;

    String cmd = e.getActionCommand();
    if ( cmd == null )
    {
    }
    else if ( cmd.equals( "start" ) )
    {

        startButton.setEnabled(false);
        stopButton.setEnabled(true);
        pauseButton.setEnabled(true);
        resumeButton.setEnabled(false);
        updatePreferences();
        openStream();
        Date tod = new Date();
        loggingStartTime = tod.getTime();
        progressBar.setIndeterminate(true);
      
    }
    else if ( cmd.equals( "stop" ) )
    {
        startButton.setEnabled(true);
        stopButton.setEnabled(false);
        pauseButton.setEnabled(false);
        resumeButton.setEnabled(false);
        updatePreferences();
        closeStream();
        progressBar.setIndeterminate(false);
    }
    else if ( cmd.equals( "pause" ) )
    {
        startButton.setEnabled(false);
        stopButton.setEnabled(true);
        pauseButton.setEnabled(false);
        resumeButton.setEnabled(true);
        progressBar.setIndeterminate(false);
    }
    else if ( cmd.equals( "resume" ) )
    {
        startButton.setEnabled(false);
        stopButton.setEnabled(true);
        pauseButton.setEnabled(true);
        resumeButton.setEnabled(false);
        progressBar.setIndeterminate(true);
    }
    else if ( cmd.equals( "browse" ) )
    {
	JFileChooser chooser = new JFileChooser();
	int rc = chooser.showOpenDialog(this);
        if ( rc == JFileChooser.APPROVE_OPTION )
	{
	    fileNameTF.setText(chooser.getSelectedFile().getAbsolutePath());
	}
    }
    else if ( cmd.equals( "interval" ) )
    {
      loggingInterval = Integer.parseInt(intervalTF.getText());
      prefs.putInt( "interval", loggingInterval );
    }
    return;
  }

public int getLoggingInterval(){ return loggingInterval; }

public void writeTimestamp( Date tod )
{
  if ( !progressBar.isIndeterminate() ) return;
  long timeElapsed = tod.getTime() - loggingStartTime;
  try
  {
    dataStream.write(String.format("%1$td/%1$tm/%1$tY %1$tT , %2$d ,",tod,timeElapsed));
  }
  catch ( IOException ioe )
  {
      System.out.println("Logging exception");
  }
}

public FileWriter getStream() { return dataStream; }

public void flush()
{
  try
  {
    dataStream.write("\n");
    dataStream.flush();
  }
  catch ( IOException ioe )
  {
      System.out.println("Logging exception");
  }
}

public boolean isLogging() { return progressBar.isIndeterminate(); }

public void updateCharts( Date tod, double rate, double pr )
{
  rateChart.getSeries().add(new Minute(tod), rate);
  pressureChart.getSeries().add(new Minute(tod), pr);
}

}
